import 'dart:async';
import 'package:http/http.dart' as http;
import 'dart:convert' as codec;
import 'expect.dart';
import 'response.dart';

/// Build fluent REST APIs routes
///
/// Example:
///     route('/book')
///       .query('count', '10')
///       .get.fetchList((m) => new Book.fromMap(m));
class Route {
  String _path = '';

  String _authority;

  Map<String, String> _query = {};

  Map<String, dynamic> _headers = {};

  http.BaseClient _client;

  /// Construct [Route] object with [path]
  Route([String path]) {
    if (path != null) this.path(path);
  }

  /// Set the [client] used to make HTTP requests
  Route withClient(http.BaseClient client) {
    _client = client;
    return this;
  }

  /// Set authority of the URL
  Route authority(String authority) {
    _authority = authority;
    return this;
  }

  /// Append path segments to the URL
  Route path(String path) {
    if (path.isEmpty) return this;

    if (!_path.endsWith('/') && !path.startsWith('/')) _path += '/';
    _path += path;
    return this;
  }

  /// Add query parameters
  Route query(String key, value) {
    _query[key] = value.toString();
    return this;
  }

  /// Add query parameters
  Route queries(Map<String, dynamic> value) {
    _query.addAll(value);
    return this;
  }

  /// Add headers
  Route header(String key, value) {
    _headers[key] = value;
    return this;
  }

  /// Add headers
  Route headers(Map<String, dynamic> values) {
    _headers.addAll(values);
    return this;
  }

  /// URL
  String get url {
    if (_authority == null && _query == null) {
      return _path;
    }
    StringBuffer sb = new StringBuffer();
    sb.write(_authority);
    if (!_authority.endsWith('/') && !_path.startsWith('/')) sb.write('/');
    sb.write(_path);
    if (_query == null) {
      return sb.toString();
    }
    _makeQueryParams(sb, _query);
    return sb.toString();
  }

  /// Converts to [Get] requester
  Get get get => new Get.copy(this);

  /// Converts to [Post] requester
  Post get post => new Post.copy(this);

  /// Converts to [Put] requester
  Put get put => new Put.copy(this);

  /// Converts to [Delete] requester
  Delete get delete => new Delete.copy(this);

  static void _makeQueryParams(StringBuffer sb, Map<String, dynamic> query) {
    if (query.length == 0) return;
    sb.write('?');

    void writeQuery(String key, String value, [bool isFirst = false]) {
      if (!isFirst) sb.write('&');
      sb.write(Uri.encodeQueryComponent(key));
      if (value != null && value.isNotEmpty) {
        sb.write("=");
        sb.write(Uri.encodeQueryComponent(value));
      }
    }

    void writeQueries(String key, value, [bool isFirst = false]) {
      if (value == null || value is String) {
        writeQuery(key, value, isFirst);
        return;
      }
      if (value is Iterable<String>) {
        for (int i = 0; i < value.length; i++) {
          if (i == 0) {
            writeQuery(key, value.elementAt(i), isFirst);
          } else {
            writeQuery(key, value.elementAt(i), false);
          }
        }
        return;
      }
    }

    for (int i = 0; i < query.length; i++) {
      String key = query.keys.elementAt(i);
      if (i == 0) {
        writeQueries(key, query[key], true);
      } else {
        writeQueries(key, query[key], false);
      }
    }
  }
}

/// Build fluent REST GET APIs
///
/// Example:
///     get('/book')
///       .query('count', '10')
///       .fetchList((m) => new Book.fromMap(m));
class Get extends Route {
  Get(String path) : super(path);

  Get.copy(Route route) {
    _authority = route._authority;
    _path = route._path;
    _query = route._query;
    _headers = route._headers;
    _client = route._client;
  }

  Get withClient(http.BaseClient client) => super.withClient(client);

  Get authority(String authority) => super.authority(authority);

  Get path(String path) => super.path(path);

  Get query(String key, value) => super.query(key, value);

  Get queries(Map<String, dynamic> value) => super.queries(value);

  Get header(String key, value) => super.header(key, value);

  Get headers(Map<String, dynamic> values) => super.headers(values);

  AsyncStringResponse get fetchResponse => new AsyncStringResponse.from(
      _client?.get(url, headers: _headers) ?? http.get(url, headers: _headers));

  Future<T> fetch<T, F>([T convert(F d)]) async {
    StringResponse resp = await fetchResponse;
    if (resp.statusCode == 200) {
      final d = codec.json.decode(resp.body);
      if (convert != null) return convert(d);
      return d;
    }
    throw resp;
  }

  Future<List<T>> fetchList<T, F>([T convert(F d)]) async {
    List d = await fetch<List, F>();
    if (convert != null) return d.cast<F>().map(convert).cast<T>().toList();
    return d;
  }

  AsyncStringResponse expect(List<Checker<Response>> conditions) =>
      fetchResponse..expect(conditions);

  AsyncStringResponse exact(
          {int statusCode,
          String body,
          String mimeType,
          String encoding,
          Map<String, String> headers,
          int contentLength}) =>
      fetchResponse
        ..exact(
            statusCode: statusCode,
            body: body,
            mimeType: mimeType,
            encoding: encoding,
            headers: headers,
            contentLength: contentLength);
}

/// Build fluent REST POST APIs
///
/// Example:
///     post('/book/${id}')
///       .json(book.toMap)
///       .fetch((m) => new Book.fromMap(m));
class Post extends Route {
  dynamic _body;

  Post(String path) : super(path);

  Post.copy(Route route) {
    _authority = route._authority;
    _path = route._path;
    _query = route._query;
    _headers = route._headers;
    _client = route._client;
  }

  Post withClient(http.BaseClient client) => super.withClient(client);

  Post authority(String authority) => super.authority(authority);

  Post path(String path) => super.path(path);

  Post query(String key, value) => super.query(key, value);

  Post queries(Map<String, dynamic> value) => super.queries(value);

  Post header(String key, value) => super.header(key, value);

  Post headers(Map<String, dynamic> values) => super.headers(values);

  Post body(String body) {
    _body = body;
    return this;
  }

  Post json(body, {bool setHeaders: true}) {
    _body = codec.json.encode(body);
    if (setHeaders) {
      header('content-type', 'application/json');
      header('Accept', 'application/json');
    }
    return this;
  }

  AsyncStringResponse get fetchResponse => new AsyncStringResponse.from(
      _client?.post(url, headers: _headers, body: _body) ??
          http.post(url, headers: _headers, body: _body));

  Future<T> fetch<T, F>([T convert(F d)]) async {
    StringResponse resp = await fetchResponse;
    if (resp.statusCode == 200) {
      final d = codec.json.decode(resp.body);
      if (convert != null) return convert(d);
      return d;
    }
    throw resp;
  }

  Future<List<T>> fetchList<T, F>([T convert(F d)]) async {
    List d = await fetch<List, F>();
    if (convert != null) return d.cast<F>().map(convert).cast<T>().toList();
    return d;
  }

  AsyncStringResponse expect(List<Checker<Response>> conditions) =>
      fetchResponse..expect(conditions);

  AsyncStringResponse exact(
          {int statusCode,
          String body,
          String mimeType,
          String encoding,
          Map<String, String> headers,
          int contentLength}) =>
      fetchResponse
        ..exact(
            statusCode: statusCode,
            body: body,
            mimeType: mimeType,
            encoding: encoding,
            headers: headers,
            contentLength: contentLength);
}

/// Build fluent REST PUT APIs
///
/// Example:
///     put('/book/${id}')
///       .json(book.toMap)
///       .fetch((m) => new Book.fromMap(m));
class Put extends Route {
  dynamic _body;

  Put(String path) : super(path);

  Put.copy(Route route) {
    _authority = route._authority;
    _path = route._path;
    _query = route._query;
    _headers = route._headers;
    _client = route._client;
  }

  Put withClient(http.BaseClient client) => super.withClient(client);

  Put authority(String authority) => super.authority(authority);

  Put path(String path) => super.path(path);

  Put query(String key, value) => super.query(key, value);

  Put queries(Map<String, dynamic> value) => super.queries(value);

  Put header(String key, value) => super.header(key, value);

  Put headers(Map<String, dynamic> values) => super.headers(values);

  Put body(String body) {
    _body = body;
    return this;
  }

  Put json(body, {bool setHeaders: true}) {
    _body = codec.json.encode(body);
    if (setHeaders) {
      header('content-type', 'application/json');
      header('Accept', 'application/json');
    }
    return this;
  }

  AsyncStringResponse get fetchResponse => new AsyncStringResponse.from(
      _client?.put(url, headers: _headers, body: _body) ??
          http.put(url, headers: _headers, body: _body));

  Future<T> fetch<T, F>([T convert(F d)]) async {
    StringResponse resp = await fetchResponse;
    if (resp.statusCode == 200) {
      final d = codec.json.decode(resp.body);
      if (convert != null) return convert(d);
      return d;
    }
    throw resp;
  }

  Future<List<T>> fetchList<T, F>([T convert(F d)]) async {
    List d = await fetch<List, F>();
    if (convert != null) return d.cast<F>().map(convert).cast<T>().toList();
    return d;
  }

  AsyncStringResponse expect(List<Checker<Response>> conditions) =>
      fetchResponse..expect(conditions);

  AsyncStringResponse exact(
          {int statusCode,
          String body,
          String mimeType,
          String encoding,
          Map<String, String> headers,
          int contentLength}) =>
      fetchResponse
        ..exact(
            statusCode: statusCode,
            body: body,
            mimeType: mimeType,
            encoding: encoding,
            headers: headers,
            contentLength: contentLength);
}

/// Build fluent REST DELETE APIs
///
/// Example:
///     delete('/book/${id}')
///       .fetchList((m) => new Book.fromMap(m));
class Delete extends Route {
  dynamic _body;

  Delete(String path) : super(path);

  Delete.copy(Route route) {
    _authority = route._authority;
    _path = route._path;
    _query = route._query;
    _headers = route._headers;
    _client = route._client;
  }

  Delete withClient(http.BaseClient client) => super.withClient(client);

  Delete authority(String authority) => super.authority(authority);

  Delete path(String path) => super.path(path);

  Delete query(String key, value) => super.query(key, value);

  Delete queries(Map<String, dynamic> value) => super.queries(value);

  Delete header(String key, value) => super.header(key, value);

  Delete headers(Map<String, dynamic> values) => super.headers(values);

  AsyncStringResponse get fetchResponse =>
      new AsyncStringResponse.from(_client?.delete(url, headers: _headers) ??
          http.delete(url, headers: _headers));

  Future<T> fetch<T, F>([T convert(F d)]) async {
    StringResponse resp = await fetchResponse;
    if (resp.statusCode == 200) {
      final d = codec.json.decode(resp.body);
      if (convert != null) return convert(d);
      return d;
    }
    throw resp;
  }

  Future<List<T>> fetchList<T, F>([T convert(F d)]) async {
    List d = await fetch<List, F>();
    if (convert != null) return d.cast<F>().map(convert).cast<T>().toList();
    return d;
  }

  AsyncStringResponse expect(List<Checker<Response>> conditions) =>
      fetchResponse..expect(conditions);

  AsyncStringResponse exact(
          {int statusCode,
          String body,
          String mimeType,
          String encoding,
          Map<String, String> headers,
          int contentLength}) =>
      fetchResponse
        ..exact(
            statusCode: statusCode,
            body: body,
            mimeType: mimeType,
            encoding: encoding,
            headers: headers,
            contentLength: contentLength);
}

Route route(String path) => new Route(path);

Get get(String path) => new Get(path);

Post post(String path) => new Post(path);

Put put(String path) => new Put(path);

Delete delete(String path) => new Delete(path);
