import 'dart:async';
import 'package:http/http.dart' as http;
import 'expect.dart';
import 'package:async/async.dart';
import 'dart:convert' as codec;

abstract class AsyncResponse<BT> {
  FutureOr<int> get statusCode;

  FutureOr<BT> get body;

  FutureOr<Map<String, String>> get headers;

  FutureOr<bool> get isRedirect;

  FutureOr<bool> get persistentConnection;

  FutureOr<String> get reasonPhrase;

  FutureOr<int> get contentLength;

  FutureOr<http.BaseRequest> get request;

  FutureOr<String> get mimeType;

  FutureOr<String> get encoding;

  AsyncResponse<BT> expect(List<Checker<Response>> conditions);

  AsyncResponse<BT> exact(
      {int statusCode,
      BT body,
      String mimeType,
      String encoding,
      Map<String, String> headers,
      int contentLength});

  FutureOr<T> json<T, F>([T convert(F d)]);

  FutureOr<List<T>> jsonList<T, F>([T convert(F d)]);
}

class AsyncTResponse<BT> extends DelegatingFuture<Response<BT>>
    implements AsyncResponse<BT> {
  AsyncTResponse(Future<Response<BT>> inner) : super(inner);

  Future<int> get statusCode => then((r) => r.statusCode);

  Future<BT> get body => then((r) => r.body);

  Future<Map<String, String>> get headers => then((r) => r.headers);

  Future<bool> get isRedirect => then((r) => r.isRedirect);

  Future<bool> get persistentConnection => then((r) => r.persistentConnection);

  Future<String> get reasonPhrase => then((r) => r.reasonPhrase);

  Future<int> get contentLength => then((r) => r.contentLength);

  Future<http.BaseRequest> get request => then((r) => r.request);

  Future<String> get mimeType => then((r) => r.mimeType);

  Future<String> get encoding => then((r) => r.encoding);

  AsyncTResponse<BT> expect(List<Checker<Response>> conditions) => then((r) {
        r.expect(conditions);
        return this;
      });

  AsyncTResponse<BT> exact(
          {int statusCode,
          BT body,
          String mimeType,
          String encoding,
          Map<String, String> headers,
          int contentLength}) =>
      then((r) {
        r.exact(
            statusCode: statusCode,
            body: body,
            mimeType: mimeType,
            encoding: encoding,
            headers: headers,
            contentLength: contentLength);
        return this;
      });

  Future<T> json<T, F>([T convert(F d)]) => then((r) => r.json(convert));

  Future<List<T>> jsonList<T, F>([T convert(F d)]) =>
      then((r) => r.jsonList(convert));
}

class AsyncStringResponse extends DelegatingFuture<Response<String>>
    implements AsyncResponse<String> {
  AsyncStringResponse(Future<Response<String>> inner) : super(inner);

  AsyncStringResponse.from(Future<http.Response> inner)
      : super(inner.then((r) => new StringResponse.from(r)));

  Future<int> get statusCode => then((r) => r.statusCode);

  Future<String> get body => then((r) => r.body);

  Future<Map<String, String>> get headers => then((r) => r.headers);

  Future<bool> get isRedirect => then((r) => r.isRedirect);

  Future<bool> get persistentConnection => then((r) => r.persistentConnection);

  Future<String> get reasonPhrase => then((r) => r.reasonPhrase);

  Future<int> get contentLength => then((r) => r.contentLength);

  Future<http.BaseRequest> get request => then((r) => r.request);

  Future<String> get mimeType => then((r) => r.mimeType);

  Future<String> get encoding => then((r) => r.encoding);

  AsyncStringResponse expect(List<Checker<Response>> conditions) =>
      then((r) => r.expect(conditions));

  AsyncStringResponse exact(
          {int statusCode,
          String body,
          String mimeType,
          String encoding,
          Map<String, String> headers,
          int contentLength}) =>
      then((r) => r.exact(
          statusCode: statusCode,
          body: body,
          mimeType: mimeType,
          encoding: encoding,
          headers: headers,
          contentLength: contentLength));

  Future<T> json<T, F>([T convert(F d)]) => then((r) => r.json(convert));

  Future<List<T>> jsonList<T, F>([T convert(F d)]) =>
      then((r) => r.jsonList(convert));
}

abstract class Response<T> implements AsyncResponse<T> {
  int get statusCode;

  T get body;

  Map<String, String> get headers;

  bool get isRedirect;

  bool get persistentConnection;

  String get reasonPhrase;

  int get contentLength;

  http.BaseRequest get request;

  String get mimeType;

  String get encoding;

  Response<T> expect(List<Checker<Response>> conditions);

  Response<T> exact(
      {int statusCode,
      T body,
      String mimeType,
      String encoding,
      Map<String, String> headers,
      int contentLength});
}

class StringResponse implements Response<String> {
  final int statusCode;

  final String body;

  final Map<String, String> headers;

  final bool isRedirect;

  final bool persistentConnection;

  final String reasonPhrase;

  final int contentLength;

  final http.BaseRequest request;

  final String mimeType;

  final String encoding;

  StringResponse(
      {this.statusCode,
      this.body,
      this.headers,
      this.isRedirect,
      this.persistentConnection,
      this.reasonPhrase,
      this.contentLength,
      this.request,
      this.mimeType,
      this.encoding});

  factory StringResponse.from(http.Response resp) => new StringResponse(
      statusCode: resp.statusCode,
      body: resp.body,
      headers: resp.headers,
      isRedirect: resp.isRedirect,
      persistentConnection: resp.persistentConnection,
      reasonPhrase: resp.reasonPhrase,
      contentLength: resp.contentLength,
      request: resp.request,
      mimeType: _getMimeType(resp.headers['content-type']),
      encoding: _getEncoding(resp.headers['content-type']));

  T json<T, F>([T convert(F d)]) {
    final d = codec.json.decode(body);
    if (convert == null) return d;
    return convert(d);
  }

  Future<List<T>> jsonList<T, F>([T convert(F d)]) async {
    List d = codec.json.decode(body);
    if (convert != null) return d.cast<F>().map(convert).cast<T>().toList();
    return d;
  }

  StringResponse expect(List<Checker<Response>> conditions) {
    final mismatches = conditions
        .map((c) => c(this))
        .reduce((List<Mismatch> v, List<Mismatch> e) => v..addAll(e))
        .toList();
    if (mismatches.length != 0) throw mismatches;
    return this;
  }

  StringResponse exact(
      {int statusCode,
      String body,
      String mimeType,
      String encoding,
      Map<String, String> headers,
      int contentLength}) {
    final conditions = <Checker<Response>>[];
    if (statusCode != null) conditions.add(statusCodeIs(statusCode));
    if (body != null) conditions.add(bodyIs(body));
    if (mimeType != null) conditions.add(mimeTypeIs(mimeType));
    if (encoding != null) conditions.add(encodingIs(encoding));
    if (headers != null) {
      headers.forEach(
          (String key, String value) => conditions.add(headersHas(key, value)));
    }
    return expect(conditions);
  }
}

String _getMimeType(String header) {
  header ??= '';
  List<String> parts = header.split(';');
  if (parts.length == 0) return 'text/plain';
  return parts.first.trim();
}

String _getEncoding(String header) {
  header ??= '';
  List<String> parts = header.split(';');
  if (parts.length < 2) return 'utf-8';
  parts.removeAt(0);
  return parts.map((s) {
    List<String> parts = s.split('=');
    if (parts.length != 2) return null;
    if (parts.first != 'charset') return null;
    return parts[1];
  }).firstWhere((e) => e != null, orElse: () => 'utf-8');
}
